(*
   Firebird Library
   File:FBLScript.pas
   Copyright (c) 2002 Alessandro Batisti
   abatisti@tiscali.it
   http://web.tiscali.it/fblib

  * This file is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This file is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
*)
{$I fbl.inc}
unit FBLScript;

interface

uses
  Classes,
  SysUtils;

type
  TStatementType = (stUnknow, stSetTerm, stSetGenerator, stSelect, stInsert, stUpdate,
    stDelete, stDDL, stCommit, stRollback, stExecProc,
    stSec, stCreateDatabase);

  TFBLScript = class(TComponent)
  private
    FScript: TStrings;
    FTerm: string;
    FEScript: string;
    FOpen: boolean;
    FEOF: boolean;
    FStatementType: TStatementType;
    procedure SetScript(Value: TStrings);
    function RemoveComment(ASql: string): string;
    function GetStatement: string;
    function GetTerminator: string;
    procedure ParseStatement(ASql: string);
    procedure Init;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Reset;
    property Statement: string read GetStatement;
    property EOF: boolean read FEOF;
    property StatementType: TStatementType read FStatementType;
  published
    property SQLScript: TStrings read FScript write SetScript;
    property Terminator: string read GetTerminator;
  end;


implementation

uses FBLConst;

constructor TFBLScript.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FScript := TStringList.Create;
  FTerm := ';';
  FOpen := False;
  //FPos:=0;
end;

destructor TFBLScript.Destroy;
begin
  FScript.Free;
  inherited Destroy;
end;

procedure TFBLScript.SetScript(Value: TStrings);
begin
  FScript.Assign(Value);
end;

function TFBLScript.GetTerminator;
begin
  Result := FTerm;
end;

procedure TFBLScript.Reset;
begin
  FEOF := False;
  FOpen := False;
end;

//------------------------------------------------------------------------------

function TFBLScript.RemoveComment(ASql: string): string;
var
  PosBegin, PosEnd: integer;
begin
  repeat
    PosBegin := Pos('/*', ASql);
    PosEnd := Pos('*/', ASql);
    if (PosBegin > 0) and (PosEnd = 0) then
      raise Exception.CreateFmt(E_SCRIPT_COMMENT, [FEScript])
    else if (PosBegin = 0) and (PosEnd > 0) then
      raise Exception.CreateFmt(E_SCRIPT_COMMENT, [FEScript])
    else
      Delete(ASql, PosBegin, PosEnd - PosBegin + 2);
  until (PosBegin = 0) and (PosEnd = 0);
  Result := ASql;
end;

//------------------------------------------------------------------------------

procedure TFBLScript.Init;
begin
  FEScript := FScript.Text;
  //RemoveComment;
  FEOF := False;
  FOpen := True;
end;

//------------------------------------------------------------------------------

function TFBLScript.GetStatement: string;
var
  Stm, UnComment: string;
  PosTerm, PosComBegin, PosComEnd: integer;
label
  check_comment;
begin
  Result := '';
  if not FOpen then Init;
  //check comment
  check_comment:
  PosComBegin := Pos('/*', FEScript);    //Pos Begin Comment
  PosComEnd := Pos('*/', FEScript);      //Pos End Comment
  PosTerm := Pos(FTerm, FEScript);       //Pos Term
  if (PosTerm < PosComEnd) and (PosComBegin < pi) then
  begin
    Delete(FEScript, PosComBegin, PosComEnd - PosComBegin + 2);
    goto check_comment;
  end;
  // Fine check comment
  if PosTerm > 0 then
  begin
    Stm := Copy(FESCript, 1,PosTerm - 1);
    Delete(FEScript, 1,PosTerm + Length(FTerm) - 1);
  end
  else
  begin
    Stm := Trim(FEScript);
    FEScript := '';
  end;
  UnComment := RemoveComment(Stm);
  if Length(Trim(FEScript)) < 3 then
    FEOF := True;
  ParseStatement(Trim(UnComment));
  if FStatementType = stDDL then
    Result := Trim(Stm)
  else
    Result := Trim(UnComment);
end;

//------------------------------------------------------------------------------

procedure TFBLScript.ParseStatement(ASql: string);
var
  Tokens: TStringList;
  CurrentToken: string;
  Idx, Len, nt: integer;
begin
  Len := Length(ASql);
  if Len = 0 then exit;
  Tokens := TStringList.Create;
  CurrentToken := '';
  Idx := 1;
  try
    (* tokenize command *)
    repeat
      case ASql[Idx] of
        #32:   //space
          begin
            if Length(Trim(CurrentToken)) > 0 then
            begin
              Tokens.Add(UpperCase(CurrentToken));
              CurrentToken := '';
            end;
          end;
        #10, #13:;
        else
          begin
            CurrentToken := CurrentToken + ASql[Idx];
            if Idx = Len then
            begin
              Tokens.Add(UpperCase(CurrentToken));
              CurrentToken := '';
            end;
          end;
      end;
      Inc(Idx);
    until Idx > Len;

    //FTokens:=Tokens.Text;
    nt := Tokens.Count;
    FStatementType := stUnknow;
    if nt > 0 then
    begin
      if Tokens.Strings[0] = 'SET' then
      begin
        if nt > 1 then
          if Tokens.Strings[1] = 'TERM' then
            if nt > 2 then
            begin
              FTerm := Tokens.Strings[2];
              FStatementType := stSetTerm;
            end
            else if Tokens.Strings[1] = 'GENERATOR' then
              FStatementType := stSetGenerator;
      end;
      if Tokens.Strings[0] = 'SELECT' then
        FStatementType := stSelect;
      if Tokens.Strings[0] = 'UPDATE' then
        FStatementType := stUpdate;
      if Tokens.Strings[0] = 'INSERT' then
        FStatementType := stInsert;
      if Tokens.Strings[0] = 'DELETE' then
        FStatementType := stDelete;
      if Tokens.Strings[0] = 'EXECUTE' then
        FStatementType := stExecProc;
      if (Tokens.Strings[0] = 'CREATE') or (Tokens.Strings[0] = 'RECREATE') or
        (Tokens.Strings[0] = 'DECLARE') or
        (Tokens.Strings[0] = 'ALTER') or (Tokens.Strings[0] = 'DROP') then
        FStatementType := stDDL;
      if Tokens.Strings[0] = 'COMMIT' then
        FStatementType := stCommit;
      if Tokens.Strings[0] = 'ROLLBACK' then
        FStatementType := stRollBack;
      if (Tokens.Strings[0] = 'GRANT') or (Tokens.Strings[0] = 'REVOKE') then
        FStatementType := stSec;
      if (Tokens.Strings[0] = 'CREATE') then
        if nt > 1 then
          if (Tokens.Strings[1] = 'DATABASE') then
            FStatementType := stCreateDatabase;
    end;

  finally
    Tokens.Free;
  end;
end;


end.
