(*
   Firebird Library
   File:FBLTransaction.pas
   Copyright (c) 2002 Alessandro Batisti
   abatisti@tiscali.it
   http://web.tiscali.it/fblib

  * This file is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This file is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
*)
{$I fbl.inc}
unit FBLTransaction;

interface

uses
  SysUtils,
 {$IFDEF D6P}
  Types,
 {$ENDIF}
  Classes,
  ibase_h, FBLDatabase;



type
  IFBLTranEvent = interface
    ['{ABAEBCBF-722A-4570-A3C9-A0F66045C6BB}']
    procedure DoOnEndTransaction;
  end;

  TTRAction = (TARollback, TACommit, TACommitRetaining);
  TTRAccessMode = (amWrite, amRead);
  TTRIsolationLevel = (ilConcurrency, ilConsistency, ilReadCommitted_rec_version,
    ilReadCommitted_no_rec_version);
  TTRLockResolution = (lrWait, lrNoWait);
  TEndTransaction = procedure(Sender: TObject; Action: TTRAction) of object;

  TFBLTransaction = class(TComponent, IFBLDbEvent)
  private
    FLibraryLoaded: boolean;
    FTRHandle: TISC_TR_HANDLE;
    FDatabase: TFBLDatabase;
    FAccessMode: TTRAccessMode;
    FIsolationLevel: TTRIsolationLevel;
    FLockResolution: TTRLockResolution;
    FTPB: PChar;
    FDSqls: TList;                   //list of attach TDsqls Objects
    FOnStartTransaction: TNotifyEvent;
    FOnEndTransaction: TEndTransaction;
    function GetTRHandle: PISC_TR_HANDLE;
    function GetInTransaction: boolean;
    function GetDbHandle: PISC_DB_HANDLE;
    procedure SetDatabase(const Value: TFBLDatabase);
    procedure EndTransaction(Action: TTRAction);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {$IFDEF FPC}
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    {$ENDIF}
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure StartTransaction;
    procedure Commit;
    procedure CommitRetaining;
    procedure RollBack;
    //from interface IFBLDbEvent
    procedure DoOnDatabaseDisconnect;
    procedure AddQuery(AQuery: IFBLTranEvent);
    procedure RemQuery(AQuery: IFBLTranEvent);
    procedure CheckDbConnected;
    property TRHandle: PISC_TR_HANDLE read GetTRHandle;
    property InTransaction: boolean read GetInTransaction;
    property DBHandle: PISC_DB_HANDLE read GetDbHandle;
  published
    property Database: TFBLDatabase read FDatabase write SetDatabase;
    property OnStartTransaction: TNotifyEvent  
      read FOnStartTransaction write FOnStartTransaction;
    property OnEndTransaction: TEndTransaction 
      read FOnEndTransaction write FOnEndTransaction;
    property AccessMode: TTRAccessMode read FAccessMode write FAccessMode default amWrite;
    property IsolationLevel: TTRIsolationLevel  
      read FIsolationLevel write FIsolationLevel default ilConcurrency;
    property LockResolution: TTRLockResolution  
      read FLockResolution write FLockResolution default lrWait;
  end;




implementation

uses FBLExcept, FBLmixf, FBLConst;

constructor TFBLTRansaction.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLibraryLoaded := False;
  CheckIbSqlLoaded;
  FLibraryLoaded := True;
  CheckIBSqlLoaded;
  FTRHandle := nil;
  FTPB := nil;
  FDSqls := TList.Create;
end;

//------------------------------------------------------------------------------

destructor TFBLTransaction.Destroy;
begin
  if FLibraryLoaded then       
  begin
    Alloc_mem(FTPB, 0);
    if inTransaction then EndTransaction(TARollback);
    if Assigned(FDatabase) then FDatabase.RemoveTransaction(self);
    FDSqls.Free;
  end;
  inherited Destroy;
end;

//------------------------------------------------------------------------------
{$IFDEF FPC}
function TFBLTransaction.QueryInterface(const IID: TGUID; out Obj): HResult;
const
  E_NOINTERFACE = HResult($80004002);
begin
 if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

function TFBLTransaction._AddRef: Integer;
begin
 Result := -1;
end;

function TFBLTransaction._Release: Integer;
begin
 Result := -1;
end;
{$ENDIF}
//------------------------------------------------------------------------------


procedure TFBLTransaction.Notification(AComponent: TComponent;
  Operation: TOperation);//override;
begin
  if Operation = opRemove then 
  begin
    if AComponent = FDatabase then FDatabase := nil;
  end;
  inherited Notification(AComponent, Operation);
end;

//------------------------------------------------------------------------------

function TFBLTransaction.GetTRHandle: PISC_TR_HANDLE;
begin
  Result := @FTRhandle;
end;

//------------------------------------------------------------------------------

function TFBLTransaction.GetInTransaction: boolean;
begin
  Result := (FTRHandle <> nil);
end;

//------------------------------------------------------------------------------

function TFBLTransaction.GetDbHandle: PISC_DB_HANDLE;
begin
  if not Assigned(FDatabase) then
    FBLError(E_TR_DB_NOT_ASSIGNED);
  Result := FDatabase.DBHandle;
end;

//------------------------------------------------------------------------------

procedure TFBLTransaction.CheckDbConnected;
begin
  if not Assigned(FDatabase) then
    FBLError(E_TR_DB_NOT_ASSIGNED);
  if not FDatabase.Connected then
    FBLError(E_DB_NOACTIVE_CON);
end;

//------------------------------------------------------------------------------

procedure TFBLTransaction.SetDatabase(const Value: TFBLDatabase);
begin
  if Assigned(Value) and (Value <> FDatabase) then
    Value.AddTransaction(self);
  FDatabase := Value;
end;

//------------------------------------------------------------------------------

procedure TFBLTRansaction.AddQuery(AQuery: IFBLTranEvent);
begin
  FDSqls.Add(Pointer(AQuery));
end;

//------------------------------------------------------------------------------

procedure TFBLTransaction.RemQuery(AQuery: IFBLTranEvent);
var
  i: integer;
begin
  i := FDSqls.IndexOf(Pointer(AQuery));
  if (i > -1) then
    FDSqls.Delete(i);
end;

//------------------------------------------------------------------------------

procedure TFBLTransaction.StartTransaction;
var
  Status_vector: ISC_STATUS_VECTOR;
  teb: TISC_TEB;
  len_buffer: integer;
begin
  CheckDbConnected;
  if inTransaction then FBLError(E_TR_ACTIVE);
  if (AccessMode = amWrite) and (IsolationLevel = ilConcurrency) and
    (LockResolution = lrWait) then
  begin
    teb.db_handle := FDataBase.DBHandle;
    teb.tpb_length := 0;
    teb.tpb_address := nil;
  end
  else
  begin
    if (IsolationLevel = ilReadCommitted_rec_version) or
      (IsolationLevel = ilReadCommitted_no_rec_version) then
    begin
      alloc_mem(FTPB, 5);
      len_buffer := 5;
      FTPB[3] := char(isc_tpb_read_committed);
      if IsolationLevel = ilReadCommitted_rec_version then
        FTPB[4] := char(isc_tpb_rec_version)
      else if IsolationLevel = ilReadCommitted_no_rec_version then
        FTPB[4] := char(isc_tpb_no_rec_version);
    end
    else
    begin
      alloc_mem(FTPB, 4);
      len_buffer := 4;
    end;

    FTPB[0] := char(isc_tpb_version3);

    if AccessMode = amWrite then FTPB[1] := char(isc_tpb_write) 
    else 
      FTPB[1] := char(isc_tpb_read);

    if Isolationlevel = ilConcurrency then FTPB[2] := char(isc_tpb_concurrency)
    else if Isolationlevel = ilConsistency then FTPB[2] := char(isc_tpb_consistency);

    if LockResolution = lrWait then FTPB[3] := char(isc_tpb_wait)
    else if LockResolution = lrNoWait then  FTPB[3] := char(isc_tpb_nowait);

    teb.db_handle := FDataBase.DBHandle;
    teb.tpb_length := len_buffer;
    teb.tpb_address := FTPB;
  end;


  if isc_start_multiple(@Status_vector, @FTRHandle, 1, @teb) <> 0 then
    FBLShowError(@Status_vector);

  if Assigned(FOnStartTransaction) then
    FOnStartTransaction(self);
end;

//-------------------------------------------------------------------------------

procedure TFBLTransaction.EndTransaction(Action: TTRAction);
var
  Status_vector: ISC_STATUS_VECTOR;
  i: integer;
begin
  if not inTransaction then FBLError(E_TR_NOACTIVE);
  case Action of
    //TARollback, TACommit, TACommitRetaining
    TARollback:
      begin
        if isc_rollback_transaction(@Status_Vector, @FTRHandle) <> 0 then
          FBLShowError(@Status_Vector);
      end;
    TACommit:
      begin
        if isc_commit_transaction(@Status_vector, @FTRHandle) <> 0 then
          FBLShowError(@Status_Vector);
      end;
    TACommitRetaining:
      begin
        if isc_commit_retaining(@Status_vector, @FTRHandle) <> 0 then
          FBLShowError(@Status_Vector);
      end;
  end;
  for i := 0 to FDSqls.Count - 1 do
    IFBLTranEvent(FDSqls[i]).DoOnEndTransaction;

  if Assigned(FOnEndTransaction) then
    FOnEndTransaction(self, Action);
end;

procedure TFBLTransaction.Commit;
begin
  EndTransaction(TACommit);
end;

procedure TFBLTransaction.CommitRetaining;
begin
  EndTransaction(TACommitRetaining);
end;

procedure TFBLTransaction.RollBack;
begin
  EndTransaction(TARollback);
end;

procedure TFBLTransaction.DoOnDatabaseDisconnect;
begin
  if InTransaction then
    EndTransaction(TARollback);
end;




end.
