{
 RawDiskAccess
 -----------------------------------------------------------
 Created by Tom Honaker. Based on the Delphi Pascal (.PAS)
 file VWIN32, created by Geir Wikran. Since Geir was so kind
 as to release his code to the open-source community, Tom
 will do likewise. :-)

 unit       : RawDiskAccess
 version    : 1.0
 last update: 2004/02/20
 written for: Delphi 7+ (but should work on 6)
 written by : Tom Honaker
 e-mail     : general@e-f-w.com

 unit       : vwin32
 version    : 1.0
 last update: 1999/04/06
 written for: Delphi 3 & 4
 written by : Geir Wikran
 e-mail     : gwikran@online.no

 Please note that while RDA is made freely available, Geir's
 license terms apply to RDA as well:
 "This source code is freeware. You may use, change, and
 distribute without charge the source code as you like. This
 unit can be used freely in any commercial applications.
 However, it may not be sold as a standalone product and the
 source code may not be included in a commercial product.
 This unit is provided as is with no warrent [sic] or
 support. Make sure to read relevant information and
 documentation from Microsoft before using this unit."
 In other words, you may use RDA for any commercial or
 private project, but you must NOT include the source itself
 in any commercial project, nor can you sell the source.

 About RawDiskAccess

 RawDiskAccess is a combined, "universal" raw disk access
 provider for Delphi 7 applications. RDA expands on
 VWIN32.PAS by providing the same services for NT-kernel
 Windows versions as VWIN32.PAS does for 9x-kernel versions.
 This system automatically determines the kernel family and
 specific OS version information (down to the service pack
 number), and from there chooses and uses the proper methods
 for direct disk access. Where possible, RDA uses Win32 API
 calls to get information, so as to not invoke the Vwin32
 VxD, or open the disk directly, unnecessarily.

}
unit RawDiskAccess;

interface

uses Windows, Classes, SysUtils, StrUtils, Dialogs;{, vwin32;}
// * Note that RDA used to refer to vwin32.pas, but now
// incorporates the stuff it used to use by reference.
// Thus, it's got all the goods in a single PAS file.

type
  // Media identifier, which holds the volume name, serial
  // number, and generic filesystem ID...
  TMediaIdentifier = packed record
    SerialNumber       : DWord;
    VolumeLabel        : string;
    FileSysType        : string;
  end;

  // DIOC registers, used by the INT13 calls for 9x-kernel
  // direct disk access...
  PDIOC_Registers = ^TDIOC_Registers;
  TDIOC_Registers = record
    case Byte of
    0: (EBX,EDX,ECX,EAX,EDI,ESI,Flags: DWord);
    1: (BX,BXE,DX,DXE,CX,CXE,AX,AXE,DI,DIE,SI,SIE: Word);
  end;

  // Read/write packet, used for raw sector reads/writes...
  TReadWritePacket = packed record
    StartSector: DWord;
    Sectors    : Word;
    Buffer     : Pointer;
  end;

  // Disk sizing info, for the particulars on a drive's sector
  // and cluster sizes, and free and total counts...
  TDiskSizingInformation = packed record
    ClusterSize: Integer;
    SectorSize: Integer;
    SectorsPerCluster: Integer;
    TotalClusters: Cardinal;
    FreeClusters: Cardinal;
    TotalSpace: Cardinal;
    FreeSpace: Cardinal;
  end;

  // Windows version information.
  TWinVersion = record
    WinName: string;
    KernelType: string;
    ServicePack: string;
    MajorVer: Integer;
    MinorVer: Integer;
    BuildNum: Integer;
  end;

  // OSVersionInfoEx declare, which wasn't included in this form in System.pas...
  POSVersionInfoEx = ^TOSVersionInfoEx;
  TOSVersionInfoEx = record
    dwOSVersionInfoSize: DWORD;
    dwMajorVersion: DWORD;
    dwMinorVersion: DWORD;
    dwBuildNumber: DWORD;
    dwPlatformId: DWORD;
    szCSDVersion: array [0..127] of Char;
    wServicePackMajor: Word;
    wServicePackMinor: Word;
    wSuiteMask: Word;
    wProductType: Byte;
    wReserved: Byte;
  end;

  // NTFS boot sector contents
  TNTFSBootRecord = packed record
    BootCode: array [0..2] of Byte;
    OemName: array [0..7] of Char;
    BytesPerSec: Word;
    SecPerClust: Byte;
    ReservedSectors: Word;
    Unused0: array [0..4] of Byte;
    MediaDescriptor: Byte;
    Unused1: Word;
    SectorsPerTrack: Word;
    NumOfHeads: Word;
    HiddenSectors: Cardinal;
    Unused2: array [0..7] of Byte;
    SectorSize: Int64;
    MTFClust: Int64;
    MTFMirrotClust: Int64;
    ClustersPerFileRec: Cardinal;
    ClustersPerIndexBlock: Cardinal;
    SerialNumber: Int64;
    Checksum: Cardinal;
    Code: array [0..425] of Byte;
    BootSignature: Word;
  end;

  // FAT32 boot sector contents
  TFAT32BootRecord = packed record
    BootCode: array [0..2] of Byte;
    OemName: array [0..7] of Char;
    BytesPerSec: Word;
    SecPerClust: Byte;
    ReservedSectors: Word;
    NumOfFatCopies: Byte;
    MaxRootEntries: Word;
    SectorsOld: Word;
    MediaDescriptor: Byte;
    SectorsPerFatOld: Word;
    SectorsPerTrack: Word;
    NumOfHeads: Word;
    HiddenSectors: Cardinal;
    SectorsSize: Cardinal;
    SectorsPerFat: Cardinal;
    Flags: Word;
    Version: Word;
    RootDirStartClust: Cardinal;
    FileSystemInfoSec: Word;
    BackupSec: Word;
    Reserved: array [0..11] of Byte;
    DriveNumber: Word;
    ExtraSignature: Byte;
    SerialNumber: Cardinal;
    VolumeName: array [0..10] of Char;
    FATName: array [0..7] of Char;
    Code: array [0..419] of Byte;
    BootSignature: Word;
  end;


  // FAT file attribute flags
  TFAT32Attributes = set of (aReadOnly, aHidden, aSystem,
                             aLabel, aDirectory, aArchive,
                             aReserved1, aReserved2);

  // FAT directory entry - all of this is in 32 bytes
  TFAT32DirEntry = packed record
    FileName: array [0..7] of Char;
    FileExt: array [0..2] of Char;
    Attributes: TFAT32Attributes;
    NT: Byte;
    TenthOfCrtSec: Byte;
    TimeCreated: Word;
    DateCreated: Word;
    DateAccessed: Word;
    StartClustHigh: Word;
    Time: Word;
    Date: Word;
    StartClust: Word;
    Size: Cardinal;
  end;

  // FAT long filename entry - another 32-byte entry
  TFAT32DirEntryLFN = packed record
    Part: Byte;
    Name1: array [0..4] of WideChar;
    Attributes: Byte;
    Reserved: Byte;
    Checksum: Byte;
    Name2: array [0..5] of WideChar;
    Reserved2: Word;
    Name3: array [0..1] of WideChar;
  end;


  // The guts - the RawDiskAccess class!
  TRawDiskAccess = class
  private
    // Map a UNC path to a local drive letter...
    function MapNetworkDrive(UNCName, Password, DriveLetter, ErrorMsg: String): Integer;
    // Unmap a drive letter from a network share...
    function DisconnectNetworkDrive(DriveLetter: String; ForceFileClose: Boolean; ErrorMsg: String): Integer;
    
    // Raw NTFS sector read...
    function RDAReadNTSectors(Drive: Byte; FirstSector: DWord; SectorCount: Word; Buffer: Pointer): Boolean;
    // Raw NTFS sector write...
    function RDAWriteNTSectors(Drive: Byte; FirstSector: DWord; SectorCount: Word; Buffer: Pointer): Boolean;

    // These were taken from Geir Wikran's awesome VWIN32 code.
    // Mad props to Geir for doing the hard work to make this
    // thing go!

    // VWIN32 interface...
    function VWIN32DIOC(ControlCode: Integer; Registers: PDIOC_Registers): Boolean;
    // Raw FATxx sector read...
    function RDARead9xSectors(Drive: Byte; Sector: DWord; Count: Word; Buffer: Pointer): Boolean;
    // Raw FATxx sector write...
    function RDAWrite9xSectors(Drive: Byte; Sector: DWord; Count: Word; Buffer: Pointer;  Mode: Byte): Boolean;
    // Lock a logical FATxx volume for raw writes...
    function RDALockLogicalVolume(Drive: Byte; Level, Permission: Byte): Boolean;
    // Unlock a logical FATzz volume...
    function RDAUnlockLogicalVolume(Drive: Byte): Boolean;

    // End of Geir's code
  published
    // Get the Windows kernel family, 9x or NT...
    function RDAKernelFamily: string;
    // Get a drive's media info (filesystem, serial #, volume label)...
    function RDAGetMediaIdentifier(Drive: Char): TMediaIdentifier;
    // Get a drive's size info (cluster size, sector size, etc.)...
    function RDAFindDiskSizeInformation(lDrive: String): TDiskSizingInformation;
    // Compare the Windows version...
    function RDAWindowsVersion(Major, Minor, Build, Platform: DWord): Boolean;
    // Get details on the Windows version...
    function RDAAcquireOSVersion: TWinVersion;
    // Read entire clusters from a drive...
    function RDAReadClusters(Drive: Byte; Cluster: DWord; Count: Word; Buffer: Pointer): Boolean;
    // Write entire clusters to a drive...
    function RDAWriteClusters(Drive: Byte; Cluster: DWord; Count: Word; Buffer: Pointer): Boolean;
    // Read sectors from a drive, with auto method select by kernel type...
    function RDAReadSectors(Drive: Byte; FirstSector: DWord; SectorCount: Word; Buffer: Pointer): Boolean;
    // Write sectors to a drive, with auto method select by kernel type...
    function RDAWriteSectors(Drive: Byte; FirstSector: DWord; SectorCount: Word; Buffer: Pointer): Boolean;
  end;

var
  // Class rename, for easier utilization...
  RDA: TRawDiskAccess;

  // Error return code from VWIN32 calls...
  VWIN32Error: DWord;

  // Handle to VWIN32, once we've opened it for direct disk
  // access...
  VWIN32Device : THandle;

const
  // Constants for OS version detection...
  VER_NT_WORKSTATION                 = 1; //0x0000001
  VER_NT_DOMAIN_CONTROLLER           = 2; //0x0000002
  VER_NT_SERVER                      = 3; //0x0000003
  VER_SERVER_NT                      = $080000000;
  VER_WORKSTATION_NT                 = $040000000;
  VER_SUITE_SMALLBUSINESS            = 1; //0x00000001
  VER_SUITE_ENTERPRISE               = 2; //0x00000002
  VER_SUITE_BACKOFFICE               = 4; //0x00000004
  VER_SUITE_COMMUNICATIONS           = 8; //0x00000008
  VER_SUITE_TERMINAL                 = 16; //0x00000010
  VER_SUITE_SMALLBUSINESS_RESTRICTED = 32; //0x00000020
  VER_SUITE_EMBEDDEDNT               = 64; //0x00000040
  VER_SUITE_DATACENTER               = 128; //0x00000080
  VER_SUITE_SINGLEUSERTS             = 256; //0x00000100
  VER_SUITE_PERSONAL                 = 512; //0x00000200
  VER_SUITE_BLADE                    = 1024; //0x0000040

  // Name of DIOC device:
  VWIN32_DEVICE_NAME        = '\\.\VWIN32';

  // VWIN32 device control codes used with VWIN32DIOC function:
  VWIN32_DIOC_CLOSE         = 0; { Close the device.                           }

  VWIN32_DIOC_DOS_IOCTL     = 1; { MS-DOS device I/O control function,         }
                                 { interrupt 21h function 4400h through 4411h  }
  VWIN32_DIOC_DOS_INT25     = 2; { MS-DOS absolute disk read command,          }
                                 { interrupt 25h.                              }
  VWIN32_DIOC_DOS_INT26     = 3; { MS-DOS absolute disk write command,         }
                                 { interrupt 26h.                              }
  VWIN32_DIOC_DOS_INT13     = 4; { Low-level BIOS disk functions,              }
                                 { interrupt 13h.                              }
  VWIN32_DIOC_DOS_DRIVEINFO = 6; { MS-DOS Interrupt 21h new function 730x.     }
                                 { Supported only by Windows 95 OSR2 and later.}

  FLAG_CARRY = $00000001;

  // Write modes for WriteSector
  WRITE_MODE_UNSPECIFIED_DATA = $0000; { Other/unknown }
  WRITE_MODE_FAT_DATA         = $1000; { FAT data }
  WRITE_MODE_DIRECTORY_DATA   = $2000; { Directory data }
  WRITE_MODE_NORMAL_FILE_DATA = $3000; { Normal file data }

  // Error codes
  ERROR_NON                  = $0000; { no error                               }

  // MS-DOS/Windows error codes:
  ERROR_INVALID_FUNCTION     = $0001; { invalid function number                }
  ERROR_FILE_NOT_FOUND       = $0002; { file not found                         }
  ERROR_ACCESS_DENIED        = $0005; { specified access denied on drive       }
  ERROR_INVALID_DRIVE        = $000F; { invalid drive number                   }
  ERROR_MEDIA_NOT_LOCKED     = $00B0; { media is not locked in drive           }
  ERROR_MEDIA_LOCKED         = $00B1; { media is locked in drive               }
  ERROR_MEDIA_NOT_REMOVABLE  = $00B2; { media is not removable                 }
  ERROR_LOCKE_COUNT_EXCEEDED = $00B4; { media locke count exceeded             }
  ERROR_EJECT_REQUEST_FAILED = $00B5; { valid media eject request failed       }

  // Interrupt 13h/25h/26h error codes:
  ERROR_BAD_COMMAND          = $10001; { bad command                           }
  ERROR_BAD_ADDRESS_MARK     = $10002; { bad address mark                      }
  ERROR_WRITE_PROTECTED      = $10003; { write-protected disk                  }
  ERROR_SECTOR_NOT_FOUND     = $10004; { requested sector not found            }
  ERROR_RESET_FAILED         = $10005; { reset failed                          }
  ERROR_DISK_CHANGED         = $10006; { disk changed (floppy disk)            }
  ERROR_PARAMETER_FAILED     = $10007; { drive parameter activity failed       }
  ERROR_DMA_FAILURE          = $10008; { DMA failure/overrun                   }
  ERROR_DMA_SEGMENT_FAULT    = $10009; { attempted DMA across 64K boundary     }
  ERROR_BAD_SECTOR_DETECTED  = $1000A; { bad sector detected                   }
  ERROR_BAD_TRACK_DETECTED   = $1000B; { bad track detected                    }
  ERROR_INVALID_MEDIA        = $1000C; { invalid media or unsupported track    }
  ERROR_INVALID_SECTORS      = $1000D; { invalid number of sectors on format   }
  ERROR_CONTROL_DATA         = $1000E; { control data address mark detected    }
  ERROR_DMA_ARBITRATION      = $1000F; { DMA arbitration level out of range    }
  ERROR_DATA_ERROR           = $10010; { data error (uncorrectable CRC or ECC) }
  ERROR_DATA_ECC_CORRECTED   = $10011; { data ECC corrected                    }
  ERROR_CONTROLLER_FAILED    = $10020; { controller failed                     }
  ERROR_SEEK_FAILED          = $10040; { seek operation failed                 }
  ERROR_DEVICE_FAILED        = $10080; { device failed to respond (timeout)    }
  ERROR_DRIVE_NOT_READY      = $100AA; { drive not ready                       }
  ERROR_UNDEFINED            = $100BB; { undefined error                       }
  ERROR_WRITE_FAULT          = $100CC; { write fault                           }
  ERROR_STATUS_REGISTER      = $100E0; { status register error                 }
  ERROR_SENSE_FAILED         = $100FF; { sense operation failed                }

  // VWIN32 custom error codes:
  ERROR_UNKNOWN              = $00100000; { unknown error                      }
  ERROR_OS_NOT_SUPPORTED     = $00200000; { OS version not supported           }
  ERROR_OPENING_DEVICE       = $00300000; { error trying to open device        }

  { Lock permission codes: }
  LOCK_ALLOW_WRITING  = $0001; { Allow write operations in level 1 lock. Write }
                               { operations are always blocked in level 2 & 3  }
                               { lock.                                         }
  LOCK_BLOCK_MAPPING  = $0002; { Block new file mappings in level 1 & 2 lock.  }
                               { New file mappings are always blocked in level }
                               { 3 lock.                                       }
                               { Read operations are always allowed in level   }
                               { 1 & 2 lock, and blocked in level 3 lock.      }
  LOCK_FOR_FORMATTING = $0004; { Locks the volume for formatting. Specified    }
                               { when a level 0 lock is obtained for the second}
                               { time.                                         }

function GetExtendedVersionEx(lpVersionInformation: POSVersionInfoEx): BOOL; external kernel32 name 'GetVersionExA';

implementation

//==========================================================

// KernelFamily
// ------------
// Returns the kernel family for the current Windows version.
// This is necessary for ensuring the use of the proper
// methods for direct disk access as the kernel type tells
// which method to use.
function TRawDiskAccess.RDAKernelFamily: string;
var
  OSVERSIONINFO: _OSVERSIONINFO;

begin
  OSVERSIONINFO.dwOSVersionInfoSize := SizeOf(OSVERSIONINFO);
  GetVersionEx(OSVERSIONINFO);

  case OSVERSIONINFO.dwPlatformId  of
    VER_PLATFORM_WIN32s: Result := 'Win32S';
    VER_PLATFORM_WIN32_WINDOWS: Result := '9x';
    VER_PLATFORM_WIN32_NT: Result := 'NT';
  end;
end;

//==========================================================

// GetMediaIdentifier
// ------------------
// Gets information on the drive, spcifically returning the
// drive's serial #, filesystem, and volume name.
// NOTE: Drive represents the drive letter, and you only need
// to provide the letter itself without colon/backslash.
function TRawDiskAccess.RDAGetMediaIdentifier(Drive: Char): TMediaIdentifier;
var
  TempBool: Boolean;
  VolumeName: array [0..255] of Char;
  VolumeNameLength: Integer;
  SerialNumber: Cardinal;
  MaxComponentLength: Cardinal;
  FileSystemFlags: Cardinal;
  FileSystemName: array [0..255] of Char;
  FileSystemNameLength: Integer;

begin
  // Prep the variables that need prepping...
  Result.SerialNumber := 0;
  Result.VolumeLabel := '';
  Result.FileSysType := '';
  VolumeNameLength := 255;
  FileSystemNameLength := 255;
  MaxComponentLength := 0;

  // Yank the info from the drive via API call...
  GetVolumeInformation(
    PAnsiChar(Drive + ':\'),
    @VolumeName, VolumeNameLength,
    @SerialNumber, MaxComponentLength, FileSystemFlags,
    @FileSystemName, FileSystemNameLength
    );

  // Populate the result type...
  Result.FileSysType := FileSystemName;
  Result.SerialNumber := SerialNumber;
  Result.VolumeLabel := VolumeName;
end;

//==========================================================

// FindDiskSizeInformation
// -----------------------
// This function returns an entire record's worth of info on
// the chosen drive. Sector size, sectors-per-cluster count,
// even free and total space. This function provides automatic
// compensation for the discrepancies and nuances between
// the use of GetDiskFreeSpace and GetDiskFreeSpaceEx on the
// various Windows versions. This function can also accept
// UNC pathnames, and will map network path to drive letters,
// perform the proper detection, and then unmap the path.
function TRawDiskAccess.RDAFindDiskSizeInformation(lDrive: String): TDiskSizingInformation;
var
  TempStr: String;
  DriveNum: Integer;
  FC: integer;
  UnmapWhenDone: Boolean;
  Password: String;
  ErrorMsg: String;

  lAnswer: Boolean;
  lFAT32Answer: Boolean;
  lpSectorsPerCluster: Cardinal;
  lpBytesPerSector: Cardinal;
  lpNumberOfFreeClusters: Cardinal;
  lpTotalNumberOfClusters: Cardinal;
  lBytesPerCluster: Cardinal;
  lNumFreeBytes: Cardinal;
  lTotalSpace: Int64;
  FAT32NumFreeBytes: Int64;
  FAT32TotalSpace: Int64;
  FAT32FreeBytesAvailable: Int64;
  DiskBytePerClusterCount: Int64;
  ClusterSize: Integer;
  ErrorRet: Cardinal;

  RootPtr: PChar;

  Kernel: THandle;

  label RetryMap;

begin
  // Prep a ton of variables...
  lAnswer := false;
  lFAT32Answer := false;
  lpSectorsPerCluster := 0;
  lpBytesPerSector := 0;
  lpNumberOfFreeClusters := 0;
  lpTotalNumberOfClusters := 0;
  lBytesPerCluster := 0;
  lNumFreeBytes := 0;
  lTotalSpace := 0;
  FAT32NumFreeBytes := 0;
  FAT32TotalSpace := 0;
  FAT32FreeBytesAvailable := 0;
  DiskBytePerClusterCount := 0;
  ClusterSize := 0;
  ErrorRet := 0;
  Password := '';
  UnmapWhenDone := false;

  // Drop if the function's not called with some sort of drive
  // letter or UNC path...
  if (lDrive = '') then Exit;

  // Tack on the colon and backslash if only a single letter
  // was passed to the function...
  if (Length(lDrive) = 1) then lDrive := lDrive + ':\';

  // Check for a UNC pathname (starts with "\\"), and if one
  // is present, map the path to a drive letter...
  if (LeftStr(lDrive, 2) = '\\') then
  begin
    // Loopage! This is here in case we hit a passworded
    // UNC path and need to retry with the password.
    RetryMap:

    // Find a free drive letter...
    DriveNum := 2;
    repeat
      Inc(DriveNum);
      TempStr := Chr(DriveNum + 96) + ':\';
      FC := GetDriveType(PAnsiChar(TempStr));
    until (FC = DRIVE_NO_ROOT_DIR) or (DriveNum = 26);

    // Found a free drive letter to map to, so let's map!
    if (FC = DRIVE_NO_ROOT_DIR) then
    begin
      FC := MapNetworkDrive(lDrive, Password, LeftStr(TempStr, 2), '');
      if (FC = 1) then
      begin
        // Mapping succeeded!
        lDrive := TempStr;
      end else begin
        // Mapping failed - let's deal with why...
        case FC of
          86:
          begin
            if (Password <> '') then
            begin
              Password := InputBox('Terminus Services - Network Connection Handler', 'The network connection requires a valid password. Please enter the password for this connection. (A blank password will abort the connection.)', '');
              goto RetryMap;
            end else begin
              ErrorMsg := 'Access to ' + lDrive + ' was denied - bad or missing password.';
            end;
          end;

          5:
            ErrorMsg := 'Access to ' + lDrive + ' was denied. Please check your network settings and retry.';

          1200:
            ErrorMsg := 'Access to ' + lDrive + ' was refused due to a BAD DEVICE error. Please check your network settings and retry.';

          487:
            ErrorMsg := 'Access to ' + lDrive + ' was refused due to an invalid network path. Please check your network settings and retry.';

          87:
            ErrorMsg := 'Access to ' + lDrive + ' was refused due to an invalid local device name. Please check your computer''s configuration and retry.';

          59:
            ErrorMsg := 'Access to ' + lDrive + ' was refused due to an unexpected network error. Please check your network settings and retry.';
        end;
      end;
    end;
  end;

  // Prep the drive name, in a form usable for API calls...
  if (LeftStr(lDrive, 2) <> '\\') then lDrive := LeftStr(lDrive, 3);
  RootPtr := PChar(lDrive + #0);

  // Since Win95 original doesn't even have the extended form of the
  // GetDiskFreeSpace API call, the best way to check whether the
  // extended call is available is to probe Kernel32 for its entry
  // point. No entry point means no extended call, so fall back to
  // the original form.
  Kernel := GetModuleHandle(Windows.Kernel32);
  if Kernel <> 0 then
    @GetDiskFreeSpaceEx := GetProcAddress(Kernel, 'GetDiskFreeSpaceExA');
  if not Assigned(GetDiskFreeSpaceEx) then
  begin
    // Windows 95 (original and SP1) versions need a simpler
    // cluster check, as these Win versions don't support FAT32
    // like Win95 OSR 2+, Win98, WinME, and the NT-kernel
    // releases.
    lAnswer := GetDiskFreeSpaceA(RootPtr,
      lpSectorsPerCluster,
      lpBytesPerSector,
      lpNumberOfFreeClusters,
      lpTotalNumberOfClusters);

    // Do some math...
    lBytesPerCluster := lpSectorsPerCluster * lpBytesPerSector;
    lNumFreeBytes := lBytesPerCluster * lpNumberOfFreeClusters;
    lTotalSpace := lpSectorsPerCluster * lpBytesPerSector * lpTotalNumberOfClusters;

    // Populate the result type...
    Result.ClusterSize := lBytesPerCluster;
    Result.SectorSize := lpBytesPerSector;
    Result.SectorsPerCluster := lpSectorsPerCluster;
    Result.TotalClusters := lpTotalNumberOfClusters;
    Result.FreeClusters := lpNumberOfFreeClusters;
    Result.TotalSpace := lTotalSpace;
    Result.FreeSpace := lNumFreeBytes;
  end else begin
    // Non-Win95-original versions use a different call that
    // supports FAT32. The 9x-kernel versions require some
    // math to deal with inaccurate API call results, but
    // NT-kernel versions are accurate.

    lFAT32Answer := GetDiskFreeSpaceExA
      (RootPtr,
      FAT32NumFreeBytes,
      FAT32TotalSpace,
      @FAT32FreeBytesAvailable);

    lTotalSpace := FAT32TotalSpace;

    // Note that we have to use both of the free-space API
    // calls so that we get all the info we need...
    lAnswer := GetDiskFreeSpaceA(RootPtr, lpSectorsPerCluster, lpBytesPerSector,
      lpNumberOfFreeClusters, lpTotalNumberOfClusters);

    // Partially populate the result type...
    Result.SectorSize := lpBytesPerSector;
    Result.SectorsPerCluster := lpSectorsPerCluster;
    Result.TotalClusters := lpTotalNumberOfClusters;
    Result.FreeClusters := lpNumberOfFreeClusters;
    Result.TotalSpace := lTotalSpace;
    Result.FreeSpace := lNumFreeBytes;

    // If this is a 9x-kernel machine, GetDiskFreeSpace is
    // not accurate for >2GB partitions, so we test
    // accordingly.
    If (RDAKernelFamily = '9x') then begin
      If (lTotalSpace > 2147483647) Then
      begin
        // Sector count will be off by a factor of four
        // for backward compatibility, so do the math.
        lpSectorsPerCluster := Round(lpSectorsPerCluster / 4);
        Result.SectorsPerCluster := lpSectorsPerCluster;
      End;
      // If we're on a 9x-kernel machine with a <2GB
      // partition, GetDiskFreeSpace -is- accurate.
      Result.ClusterSize := lpSectorsPerCluster * lpBytesPerSector;
    End else begin
      // If this is a Win NT-kernel machine, GetDiskFreeSpace
      // is accurate, so let's use its results straight.
      Result.ClusterSize := lpSectorsPerCluster * lpBytesPerSector;
    end;
  end;
end;

//==========================================================

// MapNetworkDrive
// ---------------
// This function is used by FindDiskSizeInformation to map a
// UNC path to a local drive letter.
function TRawDiskAccess.MapNetworkDrive(UNCName, Password, DriveLetter, ErrorMsg: String): Integer;
var
  Status: Integer;
  tUNCname: String;
  tPassword: String;
  tDriveLetter: String;
  lpNetResource: NETRESOURCE;

begin
  MapNetworkDrive := 9999;
  try
    tUNCname := UNCname;
    tPassword := Password;
    tDriveLetter := DriveLetter;

    If (RightStr(tUNCname, 1) = '\') Then
      tUNCname := LeftStr(tUNCname, Length(tUNCname) - 1);

    lpNetResource.dwType := RESOURCETYPE_ANY;
    lpNetResource.lpLocalName := PAnsiChar(tDriveLetter);
    lpNetResource.lpRemoteName := PAnsiChar(tUNCname);
    lpNetResource.lpProvider := '';
    Status := WNetAddConnection2(lpNetResource, PAnsiChar(tPassword), '', Cardinal(False));

    Case Status of
      WN_SUCCESS:
        ErrorMsg := '';
      WN_NOT_SUPPORTED:
        ErrorMsg := 'Function is not supported.';
      WN_OUT_OF_MEMORY:
        ErrorMsg := 'The system is out of memory.';
      WN_NET_ERROR:
        ErrorMsg := 'An error occurred on the network.';
      WN_BAD_POINTER:
        ErrorMsg := 'The network path is invalid.';
      WN_BAD_NETNAME:
        ErrorMsg := 'Invalid network resource name.';
      WN_BAD_PASSWORD:
        ErrorMsg := 'The password is invalid.';
      WN_BAD_LOCALNAME:
        ErrorMsg := 'The local device name is invalid.';
      WN_ACCESS_DENIED:
        ErrorMsg := 'A security violation occurred.';
      WN_ALREADY_CONNECTED:
        ErrorMsg := 'This drive letter is already connected to a network drive.';
    End;

    MapNetworkDrive := Status;
  except
    on E: Exception do begin
      // Error sink!
      // Yeah yeah, this isn't the PROPER way to handle an
      // error, but since we're returning an error code anyway
      // this will stop Delphi from crapping if we encounter
      // one of the errors that'd throw a code.
    end;
  end;
end;

//==========================================================

// DisconnectNetworkDrive
// ----------------------
// This function is used by FindDiskSizeInformation to
// disconnect a UNC path from a local drive letter.
Function TRawDiskAccess.DisconnectNetworkDrive(DriveLetter: String; ForceFileClose: Boolean; ErrorMsg: String): Integer;
var
  Status: Integer;
  tDriveLetter: String;

begin
  DisconnectNetworkDrive := 9999;
  try
    tDriveLetter := DriveLetter;

    Status := WNetCancelConnection2(PAnsiChar(tDriveLetter), CONNECT_UPDATE_PROFILE, ForceFileClose);
    
    Case status of
      WN_SUCCESS:
        ErrorMsg := '';
      WN_BAD_POINTER:
        ErrorMsg := 'The network path is invalid.';
      WN_BAD_VALUE:
        ErrorMsg := 'Invalid local device name';
      WN_NET_ERROR:
        ErrorMsg := 'An error occurred on the network.';
      WN_NOT_CONNECTED:
        ErrorMsg := 'The drive is not connected';
      WN_NOT_SUPPORTED:
        ErrorMsg := 'This function is not supported';
      WN_OPEN_FILES:
        ErrorMsg := 'Files are in use on this service. Drive was not disconnected.';
      WN_OUT_OF_MEMORY:
        ErrorMsg := 'The System is Out of Memory';
    End;

    DisconnectNetworkDrive := Status;
  except
    on E: Exception do begin
      // Error sink!
      // Note: See yeah-yeah comment in the MapNetworkDrive
      // function, above...
    end;
  end;
End;

//==========================================================

// WindowsVersion
// --------------
// Geir Wikran's wrapper for quickly checking the Windows
// version in a comparative sense - pass this function the
// minimum version/build of Windows you need, and it will
// return true if the existing version is the same or newer.
function TRawDiskAccess.RDAWindowsVersion(Major, Minor, Build, Platform: DWord): Boolean;
var
  OSVersionInfo: TOSVersionInfo;

begin
  with OSVersionInfo do
    Result :=
      (dwMajorVersion >= Major) and
      (dwMinorVersion >= Minor) and
      (Word(dwBuildNumber) >= Word(Build)) and
      (dwPlatformID = Platform);
end;

//==========================================================

// VWIN32DIOC
// ----------
// Geir Wikran's wrapper for direct disk I/O through the VWIN32
// interface. It simplifies VWIN32 access, automatically loading
// and unloading the VxD as needed. This code was lifted
// verbatim from his VWIN32.PAS sourcecode.
//
// NOTE: Do not call this directly. It is invoked by other
// functions as necessary.
function TRawDiskAccess.VWIN32DIOC;
var
  BytesReturned : DWord;

  // This function was tweaked to fix a couple problems, mainly an
  // initialization error and a problem with its kernel-type check. 
  function OpenDevice: Boolean;
  begin
    VWIN32Error := ERROR_NON;
    if (VWIN32Device = INVALID_HANDLE_VALUE) or (VWIN32Device = 0) then
    begin
      if (RDAKernelFamily = '9x') then
      begin
        VWIN32Device := CreateFile(VWIN32_DEVICE_NAME,0,0,nil,0,FILE_FLAG_DELETE_ON_CLOSE,0);
        if (VWIN32Device = INVALID_HANDLE_VALUE) then VWIN32Error := ERROR_OPENING_DEVICE;
      end else begin
        VWIN32Error := ERROR_OS_NOT_SUPPORTED;
      end;
    end;
    Result := VWIN32Error = ERROR_NON;
  end;

  procedure CloseDevice;
  begin
    if (VWIN32Device <> INVALID_HANDLE_VALUE) then CloseHandle(VWIN32Device);
    VWIN32Device := INVALID_HANDLE_VALUE;
    Result := true;
  end;

begin
  VWIN32Error := ERROR_NON;
  if ControlCode = VWIN32_DIOC_CLOSE then CloseDevice
  else if OpenDevice then begin
    Result := DeviceIoControl(VWIN32Device,ControlCode,Registers,SizeOf(Registers^),
                              Registers,SizeOf(Registers^),BytesReturned,nil);
    if not result then VWIN32Error := ERROR_UNKNOWN;
  end
  else Result := false;
end;

//==========================================================

// RDAReadFATSectors
// -----------------
// Geir Wikran's sector read code. Pass it a drive (where 1 =
// A:\, 2 = B:\, and so on) along with the starting sector and
// number of sectors to copy and a pointer to a buffer, and
// it will fill the buffer with data read from the disk. Please
// note that the buffer needs to be preallocated to hold enough
// data to emcompass however many sectors' worth of data you're
// after. (USUALLY, a sector is 512 bytes, but NOT ALWAYS!
// Use the FindDiskSizeInformation function to get the bytes-
// per-sector count, and use multiples of that for the buffer
// allocation.)
//
// This function provides automatic switching to the Win95
// original or 95 OSR2 and later interrupt calls as needed.
function TRawDiskAccess.RDARead9xSectors(Drive: Byte; Sector: DWord; Count: Word; Buffer: Pointer): Boolean;
var
  Registers      : TDIOC_Registers;
  ReadWritePacket: TReadWritePacket;
  OSV: TWinVersion;

begin
  VWIN32Error := ERROR_NON;
  Result := false;
  try
    OSV := RDA.RDAAcquireOSVersion;
  except
    ;
  end;

  // Check for 9x-kernel but NOT 95 original...
  if (AnsiContainsText(OSV.WinName, '95 OSR2')) or
     (AnsiContainsText(OSV.WinName, '98')) or
     (AnsiContainsText(OSV.WinName, 'ME')) then
  begin
    ReadWritePacket.StartSector := Sector;
    ReadWritePacket.Sectors := Count;
    ReadWritePacket.Buffer := Buffer;

    Registers.EAX := $7305;
    Registers.EBX := DWord(@ReadWritePacket);
    Registers.ECX := $FFFFFFFF;
    Registers.EDX := Drive;
    Registers.ESI := $0000;
    Registers.Flags := $00000000;

    if VWIN32DIOC(VWIN32_DIOC_DOS_DRIVEINFO,@Registers) then
      if ((Registers.Flags and FLAG_CARRY) = 0) then Result := true;
  end;

  // Win95-original method...
  if (not Result) then
  begin
    ReadWritePacket.StartSector := Sector;
    ReadWritePacket.Sectors := Count;
    ReadWritePacket.Buffer := Buffer;

    Registers.EAX := Drive-1;
    Registers.EBX := DWord(@ReadWritePacket);
    Registers.ECX := $FFFFFFFF;
    Registers.Flags := $00000000;
    
    if VWIN32DIOC(VWIN32_DIOC_DOS_INT25,@Registers) then begin
      if ((Registers.Flags and FLAG_CARRY) = 0) then Result := true
      else VWIN32Error := $10000 or (Registers.AX and $00FF);
    end;
  end;
end;

//==========================================================

// RDAReadNTSectors
// ----------------
// NT-kernel sector read function. Pass it a drive (where 1 =
// A:\, 2 = B:\, and so on) along with the starting sector and
// number of sectors to copy and a pointer to a buffer, and
// it will read the disk to the buffer. Please note that
// this procedure is not filesystem-specific - NTFS and FATxx
// partitions are both touchable with the same function.
function TRawDiskAccess.RDAReadNTSectors(Drive: Byte; FirstSector: DWord; SectorCount: Word; Buffer: Pointer): Boolean;
var
  hDrive: THandle;
  DSI: TDiskSizingInformation;
  DriveRoot: string;
  BytesRead: Int64;
  ReadCount: Cardinal;

begin
  // Default to "no way in hell"...
  Result := False;

  // Get the drive's sector size info, and bail if we don't
  // get what we need...
  DSI := RDAFindDiskSizeInformation(Chr(64 + Drive));
  if (DSI.SectorSize = 0) then Exit;

  // Work out how much data we need to grab...
  ReadCount := DSI.SectorSize * SectorCount;

  // Get the filename needed to open the drive itself...
  DriveRoot := '\\.\' + Chr(64 + Drive) + ':';

  // Try to open the drive. Note that we're sharing the drive,
  // as trying to open the drive with exclusive access will
  // trip all sorts of access and permissions problems!
  hDrive := CreateFile(
    PAnsiChar(DriveRoot), GENERIC_READ,
    FILE_SHARE_READ or FILE_SHARE_WRITE,
    nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

  // If the CreateFile call failed, bail out...
  if (hDrive <> INVALID_HANDLE_VALUE) then
  begin

    // Seek to the start of the sector...
    FileSeek(hDrive, Int64(FirstSector * DSI.SectorSize), 0);

    // Try to read a sector's worth of data into the buffer...
    BytesRead := FileRead(hDrive, Buffer^, ReadCount);

    // We'll rejoice if the read count equals
    Result := (BytesRead = (DSI.SectorSize * SectorCount));

    // Close the direct drive access...
    FileClose(hDrive);
  end;
end;

//==========================================================

// RDAWrite9xSectors
// -----------------
// Geir Wikran's sector write code. Pass it a drive (where 1 =
// A:\, 2 = B:\, and so on) along with the starting sector and
// number of sectors to copy and a pointer to a buffer, and
// it will write the buffer to the disk. Please note that the
// buffer needs to be preallocated and filled with enough
// data to emcompass however many sectors' worth of data you're
// writing. (USUALLY, a sector is 512 bytes, but NOT ALWAYS!)
//
// This function provides automatic switching to the Win95
// original or 95 OSR2 and later interrupt calls as needed.
function TRawDiskAccess.RDAWrite9xSectors(Drive: Byte; Sector: DWord; Count: Word; Buffer: Pointer; Mode: Byte): Boolean;
var
  Registers      : TDIOC_Registers;
  ReadWritePacket: TReadWritePacket;
  OSV: TWinVersion;

begin
  VWIN32Error := ERROR_NON;
  Result := false;
  try
    OSV := RDA.RDAAcquireOSVersion;
  except
    ;
  end;

  // Try a volume lock on the logical volume, a requirement
  // for direct disk I/O in 9x/ME...
  if (RDALockLogicalVolume(Drive, 1, LOCK_ALLOW_WRITING)) then
  begin
    // Yay! Volume lock succeeded!

    // Check for 9x-kernel but NOT 95 original...
    if (AnsiContainsText(OSV.WinName, '95 OSR2')) or
       (AnsiContainsText(OSV.WinName, '98')) or
       (AnsiContainsText(OSV.WinName, 'ME')) then
    begin
      ReadWritePacket.StartSector := Sector;
      ReadWritePacket.Sectors := Count;
      ReadWritePacket.Buffer := Buffer;

      Registers.EAX := $7305;
      Registers.EBX := DWord(@ReadWritePacket);
      Registers.ECX := $FFFFFFFF;
      Registers.EDX := Drive;
      Registers.ESI := $0001 or (Mode and $6000);
      Registers.Flags := $00000000;
      if (VWIN32DIOC(VWIN32_DIOC_DOS_DRIVEINFO,@Registers)) then
        if ((Registers.Flags and FLAG_CARRY) = 0) then Result := true;
    end;

    // Win95-original method...
    if (not Result) then
    begin
      ReadWritePacket.StartSector := Sector;
      ReadWritePacket.Sectors := Count;
      ReadWritePacket.Buffer := Buffer;

      Registers.EAX := Drive-1;
      Registers.EBX := DWord(@ReadWritePacket);
      Registers.ECX := $FFFFFFFF;
      Registers.Flags := $00000000;
      if (VWIN32DIOC(VWIN32_DIOC_DOS_INT26,@Registers)) then
      begin
        if ((Registers.Flags and FLAG_CARRY) = 0) then Result := true
        else VWIN32Error := $10000 or (Registers.AX and $00FF);
      end;
    end;

    // Unlock the volume, now that we're done...
    RDAUnlockLogicalVolume(Drive);
  end;
end;

//==========================================================

// RDAWriteNTSectors
// -----------------
// NT-kernel sector write function. Pass it a drive (where 1 =
// A:\, 2 = B:\, and so on) along with the starting sector and
// number of sectors to write and a pointer to a buffer, and
// it will write the buffer to the disk. Please note that
// this procedure is not filesystem-specific - NTFS and FATxx
// partitions are both touchable with the same function.
function TRawDiskAccess.RDAWriteNTSectors(Drive: Byte; FirstSector: DWord; SectorCount: Word; Buffer: Pointer): Boolean;
var
  hDrive: THandle;
  DSI: TDiskSizingInformation;
  DriveRoot: string;
  BytesWritten: Int64;
  ReadCount: Cardinal;

begin
  // Default to "no way in hell"...
  Result := False;

  // Get the drive's sector size info, and bail if we don't
  // get what we need...
  DSI := RDAFindDiskSizeInformation(Chr(64 + Drive));
  if (DSI.SectorSize = 0) then Exit;

  // Work out how much data we need to grab...
  ReadCount := DSI.SectorSize * SectorCount;

  // Get the filename needed to open the drive itself...
  DriveRoot := '\\.\' + Chr(64 + Drive) + ':';

  // Try to open the drive. Note that we're sharing the drive,
  // as trying to open the drive with exclusive access will
  // trip all sorts of access and permissions problems!
  hDrive := CreateFile(
    PAnsiChar(DriveRoot), GENERIC_WRITE,
    FILE_SHARE_READ or FILE_SHARE_WRITE,
    nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

  // If the CreateFile call failed, bail out...
  if (hDrive <> INVALID_HANDLE_VALUE) then
  begin

    // Seek to the start of the sector...
    FileSeek(hDrive, Int64(FirstSector * DSI.SectorSize), 0);

    // Try to read a sector's worth of data into the buffer...
    BytesWritten := FileWrite(hDrive, Buffer^, ReadCount);

    // We'll rejoice if the read count equals
    Result := (BytesWritten = (DSI.SectorSize * SectorCount));

    // Close the direct drive access...
    FileClose(hDrive);
  end;
end;

//==========================================================

// GetOSVersion
// ------------
// This function reads the current Windows release version,
// reporting it as a string that also contains a single
// basic control code that can be searched for by other
// Terminus components.
function TRawDiskAccess.RDAAcquireOSVersion: TWinVersion;
var
    OSVERSIONINFO: _OSVERSIONINFO;
    OSVERSIONINFOEX: TOSVersionInfoEx;
    TempBool: Boolean;
    WinVersion: TWinVersion;

begin
 try
    // Hit the GetVersionEx call to load theversion info into the
    // type we'll be working with...
    OSVERSIONINFO.dwOSVersionInfoSize := SizeOf(OSVERSIONINFO);
    GetVersionEx(OSVERSIONINFO);

    // Quick-prep the major/minor version IDs...
    WinVersion.MajorVer := OSVERSIONINFO.dwMajorVersion;
    WinVersion.MinorVer := OSVERSIONINFO.dwMinorVersion;

    // Win32S on Windows 3.x - this should NEVER happen!
    if OSVERSIONINFO.dwPlatformId = VER_PLATFORM_WIN32s then
    begin
        WinVersion.WinName := 'Windows 3.x + Win32';
        WinVersion.ServicePack := '';
    end;

    // Win32 on Win9x/ME...
    if OSVERSIONINFO.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS then
    begin
        WinVersion.KernelType := '9x';

        // Get the actual OS version info...
        Case OSVERSIONINFO.dwMinorVersion of

            // Win95 versions...
            0:
            begin
                WinVersion.BuildNum := (OSVERSIONINFO.dwBuildNumber and 65535);

                If AnsiContainsText(String(OSVERSIONINFO.szCSDVersion), 'C') Then
                begin
                    WinVersion.WinName := 'Windows 95 OSR2';
                    WinVersion.ServicePack := '';
                end;
                If AnsiContainsText(String(OSVERSIONINFO.szCSDVersion), 'B') Then
                begin
                    WinVersion.WinName := 'Windows 95';
                    WinVersion.ServicePack := 'Service Pack 1';
                end;
                If (not(AnsiContainsText(String(OSVERSIONINFO.szCSDVersion), 'C')) and (not AnsiContainsText(String(OSVERSIONINFO.szCSDVersion), 'B'))) Then
                begin
                    WinVersion.WinName := 'Windows 95 Original';
                    WinVersion.ServicePack := '';
                end;
            end;

            // Win98 versions...
            10:
            begin
                WinVersion.BuildNum := (OSVERSIONINFO.dwBuildNumber and 65535);

                If AnsiContainsText(String(OSVERSIONINFO.szCSDVersion), 'A') Then
                begin
                    WinVersion.WinName := 'Windows 98 Second Edition';
                    WinVersion.ServicePack := '';
                end
                else
                begin
                    WinVersion.WinName := 'Windows 98';
                    WinVersion.ServicePack := '';
                end;
            end;

            // (GAG) WinME...
            90:
            begin
                WinVersion.BuildNum := (OSVERSIONINFO.dwBuildNumber and 65535);

                WinVersion.WinName := 'Windows ME';
                WinVersion.ServicePack := '';
            end;

        end;
    end;

    // Windows NT-kernel versions...
    if OSVERSIONINFO.dwPlatformId = VER_PLATFORM_WIN32_NT then
    begin
        WinVersion.BuildNum := OSVERSIONINFO.dwBuildNumber;
        WinVersion.ServicePack := String(OSVERSIONINFO.szCSDVersion);
        WinVersion.KernelType := 'NT';

        // Try a GetVersionEx with the extended OS version info type...
        ZeroMemory(@OSVERSIONINFOEX,SizeOf(OSVERSIONINFOEX));
        OSVERSIONINFOEX.dwOSVersionInfoSize := SizeOf(OSVERSIONINFOEX);
        TempBool := GetExtendedVersionEx(@OSVERSIONINFOEX);
        GetLastError();

        Case OSVERSIONINFO.dwMajorVersion of

            3: // WinNT 3.x
            begin
                Case OSVERSIONINFO.dwMinorVersion of
                    0: WinVersion.WinName := 'Windows NT 3.0';

                    1: WinVersion.WinName := 'Windows NT 3.1';

                    5: WinVersion.WinName := 'Windows NT 3.5';

                    51: WinVersion.WinName := 'Windows NT 3.51';
                end;
            end;

            4: // WinNT 4.0
            begin
                WinVersion.WinName := 'Windows NT 4.0';

                If TempBool then
                begin

                    if (OSVERSIONINFOEX.wProductType and VER_NT_WORKSTATION) = VER_NT_WORKSTATION then
                        WinVersion.WinName := 'Windows NT 4.0 Workstation';

                    if (OSVERSIONINFOEX.wProductType and VER_NT_SERVER) = VER_NT_SERVER then
                        WinVersion.WinName := 'Windows NT 4.0 Server';

                    if (OSVERSIONINFOEX.wSuiteMask and VER_SUITE_BACKOFFICE) = VER_SUITE_BACKOFFICE then
                        WinVersion.WinName := 'Windows NT 4.0 Server w/ BackOffice Server';

                    if (OSVERSIONINFOEX.wSuiteMask and VER_SUITE_SMALLBUSINESS) = VER_SUITE_SMALLBUSINESS then
                        WinVersion.WinName := 'Windows NT 4.0 Server w/ BackOffice Small-Business Server';

                    if (OSVERSIONINFOEX.wSuiteMask and VER_SUITE_SMALLBUSINESS_RESTRICTED) = VER_SUITE_SMALLBUSINESS_RESTRICTED then
                        WinVersion.WinName := 'Windows NT 4.0 Server w/ BackOffice Small-Business Server (Rest.)';

                end;

            end;
            
            5: // Win2000 (NT 5)
            begin
                Case OSVERSIONINFO.dwMinorVersion of
                    0:
                    begin
                        WinVersion.WinName := 'Windows 2000';

                        If TempBool then
                        begin

                            if (OSVERSIONINFOEX.wProductType and VER_NT_WORKSTATION) = VER_NT_WORKSTATION then
                                WinVersion.WinName := 'Windows 2000 Professional';

                            if (OSVERSIONINFOEX.wProductType and VER_NT_SERVER) = VER_NT_SERVER then
                                WinVersion.WinName := 'Windows 2000 Server';

                            if (OSVERSIONINFOEX.wSuiteMask and VER_SUITE_DATACENTER) = VER_SUITE_DATACENTER then
                                WinVersion.WinName := 'Windows 2000 Datacenter Server';

                            if (OSVERSIONINFOEX.wSuiteMask and VER_SUITE_ENTERPRISE) = VER_SUITE_ENTERPRISE then
                                WinVersion.WinName := 'Windows 2000 Advanced Server';
                        end;
                    end;

                    1: // WinXP (NT 5.1)
                    begin
                        WinVersion.WinName := 'Windows XP';

                        If TempBool then
                        begin

                            if (OSVERSIONINFOEX.wProductType and VER_NT_WORKSTATION) = VER_NT_WORKSTATION then
                                WinVersion.WinName := 'Windows XP Professional';

                            if (OSVERSIONINFOEX.wProductType and VER_NT_SERVER) = VER_NT_SERVER then
                                WinVersion.WinName := 'Windows 2003 Server';

                            if (OSVERSIONINFOEX.wSuiteMask and VER_SUITE_PERSONAL) = VER_SUITE_PERSONAL then
                                WinVersion.WinName := 'Windows XP Home Edition';

                            if (OSVERSIONINFOEX.wSuiteMask and VER_SUITE_ENTERPRISE) = VER_SUITE_ENTERPRISE then
                                WinVersion.WinName := 'Windows 2003 Enterprise Server';
                        end;
                    end;

                    2: // Windows 2003 series (NT 5.2)
                    begin
                        WinVersion.WinName := 'Windows Server 2003';

                        If TempBool then
                        begin

                            if (OSVERSIONINFOEX.wProductType and VER_SUITE_BLADE) = VER_SUITE_BLADE then
                                WinVersion.WinName := 'Windows Server 2003 Web Edition';

                            if (OSVERSIONINFOEX.wProductType and VER_SUITE_DATACENTER) = VER_SUITE_DATACENTER then
                                WinVersion.WinName := 'Windows Server 2003 Datacenter Edition';

                            if (OSVERSIONINFOEX.wProductType and VER_SUITE_ENTERPRISE) = VER_SUITE_ENTERPRISE then
                                WinVersion.WinName := 'Windows Server 2003 Enterprise Edition';

                        end;
                    end;

                    3..9:
                    begin
                      WinVersion.WinName := 'Future Windows Version';
                    end;
                end;
            end;

            6: // Windows NT 6, the Longhorn beta
            begin
              WinVersion.WinName := 'Windows NT 6 Beta - "Longhorn"';
            end;

            7..99:
            begin
              WinVersion.WinName := 'Future Windows Version';
            end;
        end;
    end;

    // Populate the result type, starting with the OS name
    // in a human-readable format...
    Result.WinName := WinVersion.WinName;
    if (WinVersion.ServicePack <> '') then
      Result.WinName := WinVersion.WinName + ' with ' + WinVersion.ServicePack;

    // Specific OS version...
    Result.KernelType := WinVersion.KernelType;
    Result.MajorVer := WinVersion.MajorVer;
    Result.MinorVer := WinVersion.MinorVer;
    Result.BuildNum := WinVersion.BuildNum;
    Result.ServicePack := WinVersion.ServicePack; 
 except
   ;
   // Blarg! Sink yon errorness!
   // Yeah yeah, I know, bad coding practice and all that...
   // Add an error handler if it bothers you. ;-)
 end;
end;

//==========================================================

// RDAReadClusters
// ---------------
// This function reads entire clusters from a hard drive.
function TRawDiskAccess.RDAReadClusters(Drive: Byte; Cluster: DWord; Count: Word; Buffer: Pointer): Boolean;
var
  DSI: TDiskSizingInformation;
  KernelType: string;
  StartSector: Int64;

begin
  // Default to "no way in hell"...
  Result := False;

  // Get the drive's sector and sectors-per-cluster info, and
  // bail if we don't get what we need...
  DSI := RDAFindDiskSizeInformation(Chr(64 + Drive));
  if (DSI.SectorSize = 0) or (DSI.SectorsPerCluster = 0) then Exit;

  // Get the kernel type once so we won't be fondling the API
  // call so much...
  KernelType := RDAKernelFamily;

  // Determine which sector is the cluster's first...
  StartSector := Cluster * DSI.SectorsPerCluster;

  // Grab the cluster in one big shot... Note that we have
  // to use one of the two possible cluster read functions
  // as we need to use INT 13h methods for 9x kernel and
  // CreateFile for NT kernel...
  if (KernelType = '9x') then
  begin
    // 9x-kernel Win version - it'll be a FAT12/16/32
    // partition, so we'll use the VWIN32 method...
    if (not RDARead9xSectors(Drive, StartSector, DSI.SectorsPerCluster, Buffer)) then
      Exit;
  end else begin
    // NT-kernel Win version - it could be either NTFS or
    // FATxx, but since NT-kernel Windows includes a
    // system for direct drive access it doesn't matter...
    if (not RDAReadNTSectors(Drive, StartSector, DSI.SectorsPerCluster, Buffer)) then
      Exit;
  end;

  // If we got this far everything read properly, so mark
  // the result as successful...
  Result := True;
end;

//==========================================================

// RDAReadSectors
// --------------
// This function reads multiple raw sectors. The function
// includes automatic determination of which method to use
// based on the OS kernel family.
function TRawDiskAccess.RDAReadSectors(Drive: Byte; FirstSector: DWord; SectorCount: Word; Buffer: Pointer): Boolean;
begin
  Result :=
    ((RDAKernelFamily = 'NT') and (RDAReadNTSectors(Drive, FirstSector, SectorCount, Buffer))) or
    ((RDAKernelFamily = '9x') and (RDARead9xSectors(Drive, FirstSector, SectorCount, Buffer)))
end;

//==========================================================

// RDAWriteSectors
// ---------------
// This function writes multiple raw sectors. The function
// includes automatic determination of which method to use
// based on the OS kernel family.
function TRawDiskAccess.RDAWriteSectors(Drive: Byte; FirstSector: DWord; SectorCount: Word; Buffer: Pointer): Boolean;
begin
  Result :=
    ((RDAKernelFamily = 'NT') and (RDAWriteNTSectors(Drive, FirstSector, SectorCount, Buffer))) or
    ((RDAKernelFamily = '9x') and (RDAWrite9xSectors(Drive, FirstSector, SectorCount, Buffer, 1)))
end;
//==========================================================

// RDAWriteClusters
// ----------------
// This function writes entire clusters to a drive.
function TRawDiskAccess.RDAWriteClusters(Drive: Byte; Cluster: DWord; Count: Word; Buffer: Pointer): Boolean;
var
  DSI: TDiskSizingInformation;
  KernelType: string;
  StartSector: Int64;

begin
  Result := False;

  // Get the drive's sector and sectors-per-cluster info, and
  // bail if we don't get what we need...
  DSI := RDAFindDiskSizeInformation(Chr(64 + Drive));
  if (DSI.SectorSize = 0) or (DSI.SectorsPerCluster = 0) then Exit;

  // Get the kernel type once so we won't be fondling the API
  // call so much...
  KernelType := RDAKernelFamily;

  // Determine which sector is the cluster's first...
  StartSector := Cluster * DSI.SectorsPerCluster;

  // Write the cluster in one big shot... Note that we have
  // to use one of the two possible cluster write functions
  // as we need to use INT 13h methods for 9x kernel and
  // CreateFile for NT kernel...
  if (KernelType = '9x') then
  begin
    // 9x-kernel Win version - it'll be a FAT12/16/32
    // partition, so we'll use the VWIN32 method...
    if (not RDAWrite9xSectors(Drive, StartSector, DSI.SectorsPerCluster, Buffer, 1)) then
      Exit;
  end else begin
    // NT-kernel Win version - it could be either NTFS or
    // FATxx, but since NT-kernel Windows includes a
    // system for direct drive access it doesn't matter...
    if (not RDAWriteNTSectors(Drive, StartSector, DSI.SectorsPerCluster, Buffer)) then
      Exit;
  end;

  // If we got this far everything read properly, so mark
  // the result as successful...
  Result := True;
end;

//==========================================================

// RDALockLogicalVolume
// --------------------
// This function locks a logical volume for writing, as
// required by Windows 9x/ME. NT-kernel versions don't care
// about volume locks.
function TRawDiskAccess.RDALockLogicalVolume(Drive: Byte; Level, Permission: Byte): Boolean;
var
  Registers: TDIOC_Registers;
  OSV: TWinVersion;

begin
  VWIN32Error := ERROR_NON;
  Result := false;
  if (Level > 1) then Permission := $00
  else Permission := Permission and $07;

  try
    OSV := RDA.RDAAcquireOSVersion;
  except
    ;
  end;

  // Another "9x-kernel but not 95 original" check...
  if (AnsiContainsText(OSV.WinName, '95 OSR2')) or
     (AnsiContainsText(OSV.WinName, '98')) or
     (AnsiContainsText(OSV.WinName, 'ME')) then
  begin
    Registers.EAX := $440D;
    Registers.EBX := (Level shl 8) or Drive;
    Registers.ECX := $484A;
    Registers.EDX := Permission;
    Registers.Flags := $00000000;
    if (VWIN32DIOC(VWIN32_DIOC_DOS_IOCTL,@Registers)) then
      if ((Registers.Flags and FLAG_CARRY) = 0) then Result := true;
  end;

  // Win95-original method...
  if not Result then
  begin
    Registers.EAX := $440D;
    Registers.EBX := (Level shl 8) or Drive;
    Registers.ECX := $084A;
    Registers.EDX := Permission;
    Registers.Flags := $00000000;
    if (VWIN32DIOC(VWIN32_DIOC_DOS_IOCTL,@Registers)) then begin
      if ((Registers.Flags and FLAG_CARRY) = 0) then Result := true
      else VWIN32Error := Registers.AX;
    end;
  end;
end;

// RDAUnlockLogicalVolume
// ----------------------
// This function unlocks a logical volume.
function TRawDiskAccess.RDAUnlockLogicalVolume(Drive: Byte): Boolean;
var
  Registers: TDIOC_Registers;
  OSV: TWinVersion;

begin
  VWIN32Error := ERROR_NON;
  Result := false;
  try
    OSV := RDA.RDAAcquireOSVersion;
  except
    ;
  end;

  // Another "9x-kernel but not 95 original" check...
  if (AnsiContainsText(OSV.WinName, '95 OSR2')) or
     (AnsiContainsText(OSV.WinName, '98')) or
     (AnsiContainsText(OSV.WinName, 'ME')) then
  begin
    Registers.EAX := $440D;
    Registers.EBX := Drive;
    Registers.ECX := $486A;
    Registers.Flags := $00000000;
    if (VWIN32DIOC(VWIN32_DIOC_DOS_IOCTL,@Registers)) then
      if ((Registers.Flags and FLAG_CARRY) = 0) then Result := true;
  end;

  // Win95-original method...
  if not Result then
  begin
    Registers.EAX := $440D;
    Registers.EBX := Drive;
    Registers.ECX := $086A;
    Registers.Flags := $00000000;
    if (VWIN32DIOC(VWIN32_DIOC_DOS_IOCTL,@Registers)) then
      if ((Registers.Flags and FLAG_CARRY) = 0) then Result := true
      else VWIN32Error := Registers.AX;
  end;
end;

end.
