unit RawDiskAccessComponent;
{

 Raw Disk Access VCL Frontend Component
 ======================================
 Created by Tom Honaker. Uses the Raw Disk Access unit file,
 which is based on the Delphi Pascal (.PAS) file VWIN32,
 created by Geir Wikran. Since Geir was so kind as to release
 his code to the open-source community, Tom will do likewise.
 :-)

 unit       : RawDiskAccessComponent
 version    : 1.0
 last update: 2004/03/25
 written for: Delphi 7+ (but should work on 6)
 written by : Tom Honaker
 e-mail     : general@e-f-w.com

 Please note that while RDA is made freely available, Geir's
 license terms apply to RDA - and to this component - as well:
 "This source code is freeware. You may use, change, and
 distribute without charge the source code as you like. This
 unit can be used freely in any commercial applications.
 However, it may not be sold as a standalone product and the
 source code may not be included in a commercial product.
 This unit is provided as is with no warrent [sic] or
 support. Make sure to read relevant information and
 documentation from Microsoft before using this unit."
 In other words, you may use RDA for any commercial or
 private project, but you must NOT include the source itself
 in any commercial project, nor can you sell the source.


 About RawDiskAccess

 RawDiskAccess is a combined, "universal" raw disk access
 provider for Delphi 7 applications. RDA expands on
 VWIN32.PAS by providing the same services for NT-kernel
 Windows versions as VWIN32.PAS does for 9x-kernel versions.
 This system automatically determines the kernel family and
 specific OS version information (down to the service pack
 number), and from there chooses and uses the proper methods
 for direct disk access. Where possible, RDA uses Win32 API
 calls to get information, so as to not invoke the Vwin32
 VxD, or open the disk directly, unnecessarily.

 The RDA component provides a simple frontend for "drop onto
 form, set properties, work with drive" simplicity. Using the
 RDA frontend component greatly speeds and automates most of
 the process for direct disk access, while preserving both the
 sector-level access and the any-Windows-version support of
 RDA.

 PLEASE NOTE that disk access is performed synchronously by
 this component. If you need asynchronous access, create a
 thread and invoke this control programmatically, or simply
 call RDA directly, from it.


 How To Use This Component

 The component requires that certain steps be performed in a
 particular order, in order that required structures and buffer
 space can be allocated. The order is as follows:

 1. Select a drive by setting the LogicalDrive property to its
    letter designation ('C' for the C: drive, for example).
 2. Set the number of clusters or sectors (either, not both)
    to access by setting SectorCount or ClusterCount to the
    appropriate value. The component will create its buffer
    space in reaction to changing this setting, as the RDA
    system works with entire clusters or sectors and thus its
    buffer has to be sized on cluster or sector boundaries.
 3. Set a starting cluster or sector number for the beginning
    of the read or write operation, if desired.
 4. If writing data to the drive, send the complete cluster(s)'
    or sector(s)' worth of data to the component by setting
    the DiskBufferStream or DiskBufferString properties.
    PLEASE NOTE that the buffer is probably not initialized
    with any useful data and certainly won't contain the
    original disk contents unless a read is performed first!
 5. If the status/error code returned by the StatusCode property
    is 0 (ready) or 9999 (success), trigger a read by calling
    the ReadDrive procedure, or trigger a write by calling the
    WriteDrive procedure.
 6. Watch the StatusCode property for codes 9997 if a read
    operation is in progress, or 9998 if a write operation is
    in progress. If the property returns anything below 9997
    an error has occurred, and if the return value eventually
    changes to 9999 the operation has completed successfully.
 7. If reading data from the drive, read the complete cluster(s)'
    or sector(s)' worth of data by reading the DiskBufferStream
    or DiskBufferString properties.

 Suggested order:
 1. Set the drive to work with.
 2. Set the cluster or sector count.
 3. Set the starting cluster or sector if desired, or leave
    as 0 to read the boot sector or cluster.
 4. Read the drive. This will populate the buffer with the
    current contents of the cluster(s) or sector(s).
 5. Edit the disk buffer if/as desired, but DO NOT change the
    cluster/sector starting or count values or you'll reset
    the buffer.
 6. Write to the drive if desired to commit the changes to
    disk.


 Status/Error codes

 0 - Ready. - The component has loaded and is waiting for
     something to do.
 1 - Drive not selected first. - An attempt was made to perform
     an action that has to be done after the drive is selected.
 2 - Bad drive ID or drive not responding. - the selected drive
     either doesn't exist or is not responding to queries.
 10 - Cluster number too high. - The starting cluster value
      provided is higher than the drive's cluster count.
 11 - Sector number too high. - The starting sector value
      provided is higher than the drive's sector count.
 20 - Buffer size mismatch. - The size of the memory stream or
      string passed to the component does not match the size
      expected. Set the cluster or sector count first and send
      data in (cluster count * cluster size) or (sector count *
      sector size) bytes.
 21 - Buffer size not set first. - The buffer's size has not been
      determined yet. Select a non-zero value for either cluster
      or sector count and the component will set its buffer's
      size.
 30 - No cluster or sector count provided. - An attempt was made
      to read or write but no count of clusters or sectors to
      read/write was provided. Provide this by setting either
      the ClusterCount or SectorCount properties (but not both)
      before calling ReadDrive or WriteDrive.
 5000 - Read failed. - Generic read failure message.
 5001 - Write failed. - Generic write failure message.
 9997 - Reading... - The component is busy reading data from a
        drive.
 9998 - Writing... - The component is busy writing to a drive.
 9999 - Success! - The last command completed successfully.
}

interface

uses
  Windows, SysUtils, Classes, Controls, RawDiskAccess, DesignIntf,
  DesignEditors, RawDiskAccessAbout, Forms;

type
  TRawDiskAccessComponent = class(TComponent)
  private
    cDrive: Char;
    bDrive: Byte;
    DSI: TDiskSizingInformation;
    DMI: TMediaIdentifier;
    rwFirstCluster: Cardinal;
    rwFirstSector: Cardinal;
    rwClusterCount: Cardinal;
    rwSectorCount: Cardinal;
    DiskBuffer: TMemoryStream;
    DiskBufferStr: string;
    PreviousErrorNumber: Integer;
    PreviousErrorText: string;

    procedure SetDriveLetter(tDrv: Char);
    function GetDriveLetter: Char;
    function GetLogicalDriveType: Integer;
    function GetDriveClusterSize: Integer;
    function GetDriveSectorSize: Integer;
    function GetSectorsPerCluster: Integer;
    function GetTotalClusterCount: Cardinal;
    function GetFreeClusterCount: Cardinal;
    function GetTotalSpace: Cardinal;
    function GetFreeSpace: Cardinal;
    function GetSerialNumber: Cardinal;
    function GetVolumeName: string;
    function GetFileSystem: string;
    procedure SetClusterNumber(NewVal: Cardinal);
    function GetClusterNumber: Cardinal;
    procedure SetSectorNumber(NewVal: Cardinal);
    function GetSectorNumber: Cardinal;
    procedure SetClusterCount(NewVal: Cardinal);
    function GetClusterCount: Cardinal;
    procedure SetSectorCount(NewVal: Cardinal);
    function GetSectorCount: Cardinal;
    procedure SetDiskBufferStream(NewVal: TMemoryStream);
    function GetDiskBufferStream: TMemoryStream;
    procedure SetDiskBufferString(NewVal: string);
    function GetDiskBufferString: string;
    function GetStatusCode: Integer;
    function GetStatusText: string;
  protected
    { Nothing to declare! Well, nothing protected anyway... }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
  published
    property LogicalDrive: Char read GetDriveLetter Write SetDriveLetter stored False;
    property DriveType: Integer read GetLogicalDriveType;
    property ClusterSize: Integer read GetDriveClusterSize;
    property SectorSize: Integer read GetDriveSectorSize;
    property SectorsPerCluster: Integer read GetSectorsPerCluster;
    property TotalClusters: Cardinal read GetTotalClusterCount;
    property FreeClusters: Cardinal read GetFreeClusterCount;
    property TotalDiskSpace: Cardinal read GetTotalSpace;
    property FreeDiskSpace: Cardinal read GetFreeSpace;
    property SerialNumber: Cardinal read GetSerialNumber;
    property VolumeName: string read GetVolumeName;
    property FileSystem: string read GetFileSystem;
    property ClusterNumber: Cardinal read GetClusterNumber Write SetClusterNumber stored False;
    property SectorNumber: Cardinal read GetSectorNumber Write SetSectorNumber stored False;
    property ClusterCount: Cardinal read GetClusterCount Write SetClusterCount stored False;
    property SectorCount: Cardinal read GetSectorCount Write SetSectorCount stored False;
    property DiskBufferStream: TMemoryStream read GetDiskBufferStream Write SetDiskBufferStream stored False;
    property DiskBufferString: string read GetDiskBufferString Write SetDiskBufferString stored False;
    procedure ReadDrive;
    procedure WriteDrive;
    property StatusCode: Integer read GetStatusCode;
    property StatusText: string read GetStatusText;
  end;

  // Interface for the component's editor...
  TDesigner = IDesigner;
  TFormDesigner = IDesigner;
  TRawDiskAccessEditor = Class( TComponentEditor )
    Function GetVerbCount: integer; Override;
    Function GetVerb( Index: integer ): String; Override;
    Procedure ExecuteVerb( Index: integer ); Override;
  End;


procedure Register;
Procedure ShowAboutBox(Const ComponentName: String);

implementation

//
// Component creation, destruction, and registration
//
procedure Register;
begin
  RegisterComponents('EFW Components', [TRawDiskAccessComponent]);
  RegisterComponentEditor(TRawDiskAccessComponent, TRawDiskAccessEditor);
end;

constructor TRawDiskAccessComponent.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  DiskBuffer := TMemoryStream.Create;
  DiskBufferStr := '';
  PreviousErrorNumber := 0;
  PreviousErrorText := 'Ready.';
end;

destructor TRawDiskAccessComponent.Destroy;
begin
  DiskBuffer.Free;
  SetLength(DiskBufferStr, 0);

  inherited Destroy;
end;

procedure TRawDiskAccessComponent.Loaded;
begin
 inherited Loaded;
 if not (csDesigning in ComponentState) then
 begin
 end;
end;

//
//  Component editor/about handling.
//
Procedure ShowAboutBox(Const ComponentName: String);
begin
  with TRawDiskAccessAbout.Create(Application) do
  try
    ShowModal;
  finally
    Free;
  end;
end;

Function TRawDiskAccessEditor.GetVerbCount: integer;
Begin
  Result := 1;
End;

Function TRawDiskAccessEditor.GetVerb(Index: integer): String;
Begin
  Case Index Of
    0: Result := 'About TRawDiskAccess VCL Component... (C) 2004 by EFW';
  End;
End;

Procedure TRawDiskAccessEditor.ExecuteVerb(Index: integer);
Begin
  Case Index Of
    0: ShowAboutBox('TRawDiskAccessComponent');
  End;
End;

//
// The guts of the component - the property handlers!
//

// Drive letter setting
// --------------------
// When the drive letter is set, the component obtains the
// drive's geometry.
procedure TRawDiskAccessComponent.SetDriveLetter(tDrv: Char);
begin
  cDrive := AnsiUppercase(tDrv)[1];
  bDrive := Ord(cDrive) - 64;

  if (GetDriveType(PAnsiChar(cDrive + ':\')) > 1) then
  begin
    DSI := RDA.RDAFindDiskSizeInformation(cDrive);
    DMI := RDA.RDAGetMediaIdentifier(cDrive);
  end else begin
    PreviousErrorNumber := 2;
    PreviousErrorText := 'Bad drive ID or drive not responding.';
    cDrive := #0;
    bdrive := 0;
  end;              

  PreviousErrorNumber := 9999;
  PreviousErrorText := 'Success!';
end;

function TRawDiskAccessComponent.GetDriveLetter: Char;
begin
  Result := cDrive;
end;

// Drive type
// ----------
// This property returns the result of the GetDriveType API
// call. (DRIVE_FIXED, DRIVE_REMOTE, etc.)
function TRawDiskAccessComponent.GetLogicalDriveType: Integer;
begin
  Result := GetDriveType(PAnsiChar(cDrive + ':\'));
end;

// Drive cluster size (bytes)
// --------------------------
function TRawDiskAccessComponent.GetDriveClusterSize: Integer;
begin
  Result := DSI.ClusterSize;
end;

// Drive sector size (bytes)
// -------------------------
function TRawDiskAccessComponent.GetDriveSectorSize: Integer;
begin
  Result := DSI.SectorSize;
end;

// Sector count per cluster
// ------------------------
function TRawDiskAccessComponent.GetSectorsPerCluster: Integer;
begin
  Result := DSI.SectorsPerCluster;
end;

// Total cluster count
// -------------------
function TRawDiskAccessComponent.GetTotalClusterCount: Cardinal;
begin
  Result := DSI.TotalClusters;
end;

// Free cluster count
// ------------------
function TRawDiskAccessComponent.GetFreeClusterCount: Cardinal;
begin
  Result := DSI.FreeClusters;
end;

// Total disk space
// ----------------
function TRawDiskAccessComponent.GetTotalSpace: Cardinal;
begin
  Result := DSI.TotalSpace;
end;

// Free disk space
// ---------------
function TRawDiskAccessComponent.GetFreeSpace: Cardinal;
begin
  Result := DSI.FreeSpace;
end;

// Drive's serial number
// ---------------------
function TRawDiskAccessComponent.GetSerialNumber: Cardinal;
begin
  Result := DMI.SerialNumber;
end;

// Drive's volume label (if any)
// -----------------------------
function TRawDiskAccessComponent.GetVolumeName: string;
begin
  Result := DMI.VolumeLabel;
end;

// Drive's filesystem
// ------------------
// Basic drive filesystem (generic FAT or NTFS test, does not
// discriminate between FAT types)
function TRawDiskAccessComponent.GetFileSystem: string;
begin
  Result := DMI.FileSysType;
end;

// Set or get the cluster to access
// --------------------------------
// When the cluster count or first cluster to read/write is set,
// the component will automatically allocate and clear the
// disk buffer space, and set the sector start and count values
// to zero.
procedure TRawDiskAccessComponent.SetClusterNumber(NewVal: Cardinal);
begin
  if (cDrive = #0) then
  begin
    PreviousErrorNumber := 1;
    PreviousErrorText := 'Drive not selected first.';
    Exit;
  end;

  if (NewVal > DSI.TotalClusters) then
  begin
    PreviousErrorNumber := 10;
    PreviousErrorText := 'Cluster number too high.';
    Exit;
  end;

  rwFirstCluster := NewVal;
  rwFirstSector := 0;
  rwSectorCount := 0;

  if (rwClusterCount > 0) then
  begin
    SetLength(DiskBufferStr, rwClusterCount * DSI.ClusterSize);
    DiskBuffer.Size := rwClusterCount * DSI.ClusterSize;
    SetDiskBufferString(StringOfChar(#0, Length(DiskBufferStr)));
  end;

  PreviousErrorNumber := 9999;
  PreviousErrorText := 'Success!';
end;

function TRawDiskAccessComponent.GetClusterNumber: Cardinal;
begin
  Result := rwFirstCluster;
end;

// Set or get the sector to access
// -------------------------------
// When the sector count or first sector to read/write is set,
// the component will automatically allocate and clear the
// disk buffer space, and set the cluster start and count values
// to zero.
procedure TRawDiskAccessComponent.SetSectorNumber(NewVal: Cardinal);
begin
  if (cDrive = #0) then
  begin
    PreviousErrorNumber := 1;
    PreviousErrorText := 'Drive not selected first.';
    Exit;
  end;

  if (NewVal > DSI.TotalClusters * DSI.SectorsPerCluster) then
  begin
    PreviousErrorNumber := 11;
    PreviousErrorText := 'Sector number too high.';
    Exit;
  end;

  rwFirstSector := NewVal;
  rwFirstCluster := 0;
  rwClusterCount := 0;

  if (rwSectorCount > 0) then
  begin
    SetLength(DiskBufferStr, rwSectorCount * DSI.SectorSize);
    DiskBuffer.Size := rwSectorCount * DSI.SectorSize;
    SetDiskBufferString(StringOfChar(#0, Length(DiskBufferStr)));
  end;

  PreviousErrorNumber := 9999;
  PreviousErrorText := 'Success!';
end;

function TRawDiskAccessComponent.GetSectorNumber: Cardinal;
begin
  Result := rwFirstSector;
end;

// Number of clusters to read/write
// --------------------------------
procedure TRawDiskAccessComponent.SetClusterCount(NewVal: Cardinal);
begin
  if (cDrive = #0) then
  begin
    PreviousErrorNumber := 1;
    PreviousErrorText := 'Drive not selected first.';
    Exit;
  end;

  rwClusterCount := NewVal;
  rwFirstSector := 0;
  rwSectorCount := 0;

  if (rwClusterCount > 0) then
  begin
    SetLength(DiskBufferStr, rwClusterCount * DSI.ClusterSize);
    DiskBuffer.Size := rwClusterCount * DSI.ClusterSize;
    SetDiskBufferString(StringOfChar(#0, Length(DiskBufferStr)));
  end;

  PreviousErrorNumber := 9999;
  PreviousErrorText := 'Success!';
end;

function TRawDiskAccessComponent.GetClusterCount: Cardinal;
begin
  Result := rwClusterCount;
end;

// Number of sectors to read/write
// -------------------------------
procedure TRawDiskAccessComponent.SetSectorCount(NewVal: Cardinal);
begin
  if (cDrive = #0) then
  begin
    PreviousErrorNumber := 1;
    PreviousErrorText := 'Drive not selected first.';
    Exit;
  end;

  rwSectorCount := NewVal;
  rwFirstCluster := 0;
  rwClusterCount := 0;

  if (rwSectorCount > 0) then
  begin
    SetLength(DiskBufferStr, rwSectorCount * DSI.SectorSize);
    DiskBuffer.Size := rwSectorCount * DSI.SectorSize;
    SetDiskBufferString(StringOfChar(#0, Length(DiskBufferStr)));
  end;

  PreviousErrorNumber := 9999;
  PreviousErrorText := 'Success!';
end;

function TRawDiskAccessComponent.GetSectorCount: Cardinal;
begin
  Result := rwSectorCount;
end;

// Disk buffer, in its native memory stream form
// ---------------------------------------------
procedure TRawDiskAccessComponent.SetDiskBufferStream(NewVal: TMemoryStream);
begin
  if (cDrive = #0) then
  begin
    PreviousErrorNumber := 1;
    PreviousErrorText := 'Drive not selected first.';
    Exit;
  end;

  if (DiskBuffer.Size = 0) then
  begin
    PreviousErrorNumber := 21;
    PreviousErrorText := 'Buffer size not set first.';
    Exit;
  end;

  if (NewVal.Size <> DiskBuffer.Size) then
  begin
    PreviousErrorNumber := 20;
    PreviousErrorText := 'Buffer size mismatch.';
    Exit;
  end;

  DiskBuffer.Position := 0;
  DiskBuffer := NewVal;
  SetLength(DiskBufferStr, DiskBuffer.Size);
  DiskBuffer.ReadBuffer(Pointer(DiskBufferStr)^, Length(DiskBufferStr));

  PreviousErrorNumber := 9999;
  PreviousErrorText := 'Success!';
end;

function TRawDiskAccessComponent.GetDiskBufferStream: TMemoryStream;
begin
  Result := DiskBuffer;
end;

// Disk buffer, as a string
// ------------------------
procedure TRawDiskAccessComponent.SetDiskBufferString(NewVal: string);
begin
  if (cDrive = #0) then
  begin
    PreviousErrorNumber := 1;
    PreviousErrorText := 'Drive not selected first.';
    Exit;
  end;

  if (DiskBuffer.Size = 0) then
  begin
    PreviousErrorNumber := 21;
    PreviousErrorText := 'Buffer size not set first.';
    Exit;
  end;

  if (Length(NewVal) <> Length(DiskBufferStr)) then
  begin
    PreviousErrorNumber := 20;
    PreviousErrorText := 'Buffer size mismatch.';
    Exit;
  end;

  DiskBufferStr := NewVal;
  DiskBuffer.Size := Length(DiskBufferStr);
  DiskBuffer.Position := 0;
  DiskBuffer.WriteBuffer(Pointer(DiskBufferStr)^, Length(DiskBufferStr));

  PreviousErrorNumber := 9999;
  PreviousErrorText := 'Success!';
end;

function TRawDiskAccessComponent.GetDiskBufferString: string;
begin
  Result := DiskBufferStr;
end;

// Status/error codes
// ------------------
function TRawDiskAccessComponent.GetStatusCode: Integer;
begin
  Result := PreviousErrorNumber;
end;

function TRawDiskAccessComponent.GetStatusText: string;
begin
  Result := PreviousErrorText;
end;

// Read the drive
// --------------
// This procedure triggers a RDA read of the specified cluster(s)
// or sector(s).
procedure TRawDiskAccessComponent.ReadDrive;
begin
  if (rwClusterCount = 0) and (rwSectorCount = 0) then
  begin
    PreviousErrorNumber := 30;
    PreviousErrorText := 'No cluster or sector count provided.';
    Exit;
  end;

  PreviousErrorNumber := 9997;
  PreviousErrorText := 'Reading...';

  if (rwClusterCount > 0) then
  begin
    // Read cluster(s)...
    if (RDA.RDAReadClusters(bDrive, rwFirstCluster, rwClusterCount, DiskBuffer.Memory)) then
    begin
      PreviousErrorNumber := 9999;
      PreviousErrorText := 'Success!';
    end else begin
      PreviousErrorNumber := 5000;
      PreviousErrorText := 'Read failed.';
    end;
  end else begin
    // Read sector(s)...
    if (RDA.RDAReadSectors(bDrive, rwFirstSector, rwSectorCount, DiskBuffer.Memory)) then
    begin
      PreviousErrorNumber := 9999;
      PreviousErrorText := 'Success!';
    end else begin
      PreviousErrorNumber := 5000;
      PreviousErrorText := 'Read failed.';
    end;
  end;
end;

// Write to the drive
// ------------------
// This procedure triggers a RDA write to the specified cluster(s)
// or sector(s).
procedure TRawDiskAccessComponent.WriteDrive;
begin
  if (rwClusterCount = 0) and (rwSectorCount = 0) then
  begin
    PreviousErrorNumber := 30;
    PreviousErrorText := 'No cluster or sector count provided.';
    Exit;
  end;

  PreviousErrorNumber := 9998;
  PreviousErrorText := 'Writing...';

  if (rwClusterCount > 0) then
  begin
    // Read cluster(s)...
    if (RDA.RDAWriteClusters(bDrive, rwFirstCluster, rwClusterCount, DiskBuffer.Memory)) then
    begin
      PreviousErrorNumber := 9999;
      PreviousErrorText := 'Success!';
    end else begin
      PreviousErrorNumber := 5001;
      PreviousErrorText := 'Write failed.';
    end;
  end else begin
    // Read sector(s)...
    if (RDA.RDAWriteSectors(bDrive, rwFirstSector, rwSectorCount, DiskBuffer.Memory)) then
    begin
      PreviousErrorNumber := 9999;
      PreviousErrorText := 'Success!';
    end else begin
      PreviousErrorNumber := 5001;
      PreviousErrorText := 'Write failed.';
    end;
  end;
end;

end.
